import collections
import decimal
from abc import ABCMeta, abstractmethod
from datetime import date, datetime
from decimal import Decimal
from inspect import Traceback
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    IO,
    ItemsView,
    Iterable,
    Iterator,
    KeysView,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    ValuesView,
)

import typing

from .utilities.types import NoneType, Null, UNDEFINED, Undefined


__all__: List[str] = [
    "OrderedDict",
    "Types",
    "MutableTypes",
    "Hooks",
    "ObjectHooks",
    "ArrayHooks",
    "DictionaryHooks",
    "Readable",
    "Meta",
    "ObjectMeta",
    "DictionaryMeta",
    "ArrayMeta",
    "Properties",
    "Model",
    "Dictionary",
    "Object",
    "Array",
    "Property",
    "String",
    "Date",
    "DateTime",
    "Bytes",
    "Enumerated",
    "Number",
    "Integer",
    "Boolean",
    "ArrayProperty",
    "DictionaryProperty",
    "Version",
    "MARSHALLABLE_TYPES",
    "JSON_TYPES",
    "JSONTypes",
    "MarshallableTypes",
]

if typing.TYPE_CHECKING:
    OrderedDict = collections.OrderedDict
else:
    OrderedDict = typing.Dict


def _check_methods(class_: type, methods: Iterable[str]) -> Optional[bool]:
    mro: Tuple[type, ...] = class_.__mro__
    method: str
    base_class: type
    for method in methods:
        for base_class in mro:
            if method in base_class.__dict__:
                if base_class.__dict__[method] is None:
                    return NotImplemented
                break
        else:
            return NotImplemented
    return True


class Types(metaclass=ABCMeta):

    # noinspection PyUnusedLocal,PyMissingConstructor
    @abstractmethod
    def __init__(
        self,
        items: Optional[
            Union[Iterable[Union[type, "Property"]], type, "Property"]
        ] = None,
    ) -> None:
        pass

    @abstractmethod
    def __iter__(self) -> Iterator[Union[type, "Property"]]:  # type: ignore
        pass

    @abstractmethod
    def __copy__(self) -> "Types":
        pass

    @abstractmethod
    def __deepcopy__(self, memo: dict = None) -> "Types":
        pass

    @abstractmethod
    def __repr__(self) -> str:
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def __contains__(
        self, item: Union[type, "Property"]  # type: ignore
    ) -> bool:
        pass

    def __getitem__(self, index: int) -> Union[type, "Property"]:
        pass


class MutableTypes(Types, metaclass=ABCMeta):
    """
    TODO
    """

    @abstractmethod
    def __setitem__(  # type: ignore
        self, index: int, value: Union[type, "Property"]
    ) -> None:
        pass

    @abstractmethod
    def append(self, value: Union[type, "Property"]) -> None:
        pass

    @abstractmethod
    def extend(self, values: Iterable[Union[type, "Property"]]) -> None:
        pass

    @abstractmethod
    def __delitem__(self, index: int) -> None:  # type: ignore
        pass

    @abstractmethod
    def pop(self, index: int = -1) -> Union[type, "Property"]:
        pass


class Hooks(metaclass=ABCMeta):
    """
    TODO
    """

    before_marshal: Optional[Callable[["Model"], Any]]
    after_marshal: Optional[Callable[["JSONTypes"], Any]]
    before_unmarshal: Optional[Callable[["MarshallableTypes"], Any]]
    after_unmarshal: Optional[Callable[["Model"], Any]]
    before_serialize: Optional[Callable[["JSONTypes"], Any]]
    after_serialize: Optional[Callable[[str], str]]
    before_validate: Optional[Callable[["Model"], Any]]
    after_validate: Optional[Callable[["Model"], None]]

    # noinspection PyUnusedLocal
    @abstractmethod
    def __init__(  # noqa
        self,
        before_marshal: Optional[
            Callable[["MarshallableTypes"], "MarshallableTypes"]
        ] = None,
        after_marshal: Optional[Callable[["JSONTypes"], "JSONTypes"]] = None,
        before_unmarshal: Optional[
            Callable[["MarshallableTypes"], "MarshallableTypes"]
        ] = None,
        after_unmarshal: Optional[
            Callable[["MarshallableTypes"], "MarshallableTypes"]
        ] = None,
        before_serialize: Optional[
            Callable[["JSONTypes"], "JSONTypes"]
        ] = None,
        after_serialize: Optional[Callable[[str], str]] = None,
        before_validate: Optional[Callable[["Model"], "Model"]] = None,
        after_validate: Optional[Callable[["Model"], None]] = None,
    ) -> None:
        pass

    @abstractmethod
    def __copy__(self) -> "Hooks":
        pass

    @abstractmethod
    def __deepcopy__(self, memo: dict = None) -> "Hooks":
        pass

    @abstractmethod
    def __bool__(self) -> bool:
        pass


class ObjectHooks(Hooks, metaclass=ABCMeta):
    """
    TODO
    """

    before_setattr: Optional[
        Callable[
            ["Object", str, "MarshallableTypes"],
            Tuple[str, "MarshallableTypes"],
        ]
    ]
    after_setattr: Optional[
        Callable[["Object", str, "MarshallableTypes"], None]
    ]
    before_setitem: Optional[
        Callable[["Object", str, "MarshallableTypes"], Tuple[str, Any]]
    ]
    after_setitem: Optional[
        Callable[["Object", str, "MarshallableTypes"], None]
    ]

    # noinspection PyUnusedLocal
    @abstractmethod
    def __init__(  # noqa
        self,
        before_marshal: Optional[Callable[["MarshallableTypes"], Any]] = None,
        after_marshal: Optional[Callable[["JSONTypes"], Any]] = None,
        before_unmarshal: Optional[
            Callable[["MarshallableTypes"], Any]
        ] = None,
        after_unmarshal: Optional[Callable[["MarshallableTypes"], Any]] = None,
        before_serialize: Optional[Callable[["JSONTypes"], Any]] = None,
        after_serialize: Optional[Callable[[str], str]] = None,
        before_validate: Optional[Callable[["Model"], "Model"]] = None,
        after_validate: Optional[Callable[["Model"], None]] = None,
        before_setattr: Optional[
            Callable[
                ["Object", str, "MarshallableTypes"],
                Tuple[str, "MarshallableTypes"],
            ]
        ] = None,
        after_setattr: Optional[
            Callable[["Object", str, "MarshallableTypes"], None]
        ] = None,
        before_setitem: Optional[
            Callable[
                ["Object", str, "MarshallableTypes"],
                Tuple[str, "MarshallableTypes"],
            ]
        ] = None,
        after_setitem: Optional[
            Callable[["Object", str, "MarshallableTypes"], None]
        ] = None,
    ) -> None:
        pass


class ArrayHooks(Hooks, metaclass=ABCMeta):
    """
    TODO
    """

    before_setitem: Optional[
        Callable[["Array", int, "MarshallableTypes"], Tuple[int, Any]]
    ]
    after_setitem: Optional[
        Callable[["Array", int, "MarshallableTypes"], None]
    ]
    before_append: Optional[
        Callable[["Array", "MarshallableTypes"], "MarshallableTypes"]
    ]
    after_append: Optional[Callable[["Array", "MarshallableTypes"], None]]

    # noinspection PyUnusedLocal
    @abstractmethod
    def __init__(  # noqa
        self,
        before_marshal: Optional[Callable[["MarshallableTypes"], Any]] = None,
        after_marshal: Optional[Callable[["JSONTypes"], Any]] = None,
        before_unmarshal: Optional[
            Callable[["MarshallableTypes"], Any]
        ] = None,
        after_unmarshal: Optional[Callable[["MarshallableTypes"], Any]] = None,
        before_serialize: Optional[Callable[["JSONTypes"], Any]] = None,
        after_serialize: Optional[Callable[[str], str]] = None,
        before_validate: Optional[Callable[["Model"], "Model"]] = None,
        after_validate: Optional[Callable[["Model"], None]] = None,
        before_setitem: Optional[
            Callable[["Array", int, "MarshallableTypes"], Tuple[int, Any]]
        ] = None,
        after_setitem: Optional[
            Callable[["Array", int, "MarshallableTypes"], None]
        ] = None,
        before_append: Optional[
            Callable[["Array", "MarshallableTypes"], Optional[Any]]
        ] = None,
        after_append: Optional[
            Callable[["Array", "MarshallableTypes"], None]
        ] = None,
    ) -> None:
        pass


class DictionaryHooks(Hooks, metaclass=ABCMeta):
    """
    TODO
    """

    before_setitem: Optional[
        Callable[["Dictionary", str, "MarshallableTypes"], Tuple[str, Any]]
    ]
    after_setitem: Optional[
        Callable[["Dictionary", str, "MarshallableTypes"], None]
    ]

    # noinspection PyUnusedLocal
    def __init__(  # noqa
        self,
        before_marshal: Optional[
            Callable[["MarshallableTypes"], "MarshallableTypes"]
        ] = None,
        after_marshal: Optional[Callable[["JSONTypes"], "JSONTypes"]] = None,
        before_unmarshal: Optional[
            Callable[["MarshallableTypes"], "MarshallableTypes"]
        ] = None,
        after_unmarshal: Optional[
            Callable[["MarshallableTypes"], "MarshallableTypes"]
        ] = None,
        before_serialize: Optional[Callable[["JSONTypes"], Any]] = None,
        after_serialize: Optional[Callable[[str], str]] = None,
        before_validate: Optional[Callable[["Model"], "Model"]] = None,
        after_validate: Optional[Callable[["Model"], None]] = None,
        before_setitem: Optional[
            Callable[
                ["Dictionary", str, "MarshallableTypes"],
                Tuple[str, Any],
            ]
        ] = None,
        after_setitem: Optional[
            Callable[["Dictionary", str, "MarshallableTypes"], None]
        ] = None,
    ) -> None:
        pass


class Readable(metaclass=ABCMeta):
    """
    A generic ABC for IO-like objects which are readable but not
    necessarily writable.=
    """

    register: Callable[[type], None]

    @abstractmethod
    def read(self, n: int = -1) -> Union[str, bytes]:
        pass

    @abstractmethod
    def readline(self, limit: int = -1) -> Union[str, bytes]:
        pass

    @abstractmethod
    def readlines(self, hint: int = -1) -> List[Union[str, bytes]]:
        pass

    @abstractmethod
    def seek(self, offset: int, whence: int = 0) -> int:
        pass

    @abstractmethod
    def __enter__(self) -> IO:
        pass

    @abstractmethod
    def __exit__(
        self, type_: type, value: Exception, traceback: Traceback
    ) -> None:
        pass

    @classmethod
    def __subclasshook__(cls, subclass: type) -> Optional[bool]:
        if cls is Readable:
            return _check_methods(subclass, ("read",))
        return NotImplemented


class Meta(metaclass=ABCMeta):

    pass


class ObjectMeta(Meta, metaclass=ABCMeta):
    """
    TODO
    """

    # noinspection PyUnusedLocal
    @abstractmethod
    def __init__(
        self,
        properties: Union[
            Mapping[str, "Property"],
            Iterable[Tuple[str, "Property"]],
            "Properties",
            None,
        ] = None,
    ) -> None:
        pass

    @property  # type: ignore
    @abstractmethod
    def properties(self) -> Optional["Properties"]:
        pass

    @properties.setter  # type: ignore
    @abstractmethod
    def properties(
        self,
        properties_: Union[
            Mapping[str, "Property"],
            Iterable[Tuple[str, "Property"]],
            "Properties",
            None,
        ],
    ) -> None:
        pass


class DictionaryMeta(Meta, metaclass=ABCMeta):
    """
    TODO
    """

    # noinspection PyUnusedLocal
    @abstractmethod
    def __init__(
        self,
        value_types: Union[
            Iterable[Union["Property", type]], Types, None, "Property", type
        ] = None,
    ) -> None:
        pass

    @property  # type: ignore
    @abstractmethod
    def value_types(self) -> Optional[Types]:
        pass

    @value_types.setter  # type: ignore
    @abstractmethod
    def value_types(
        self,
        value_types: Union[
            Iterable[Union["Property", type]], Types, None, "Property", type
        ],
    ) -> None:
        pass


class ArrayMeta(Meta, metaclass=ABCMeta):
    """
    TODO
    """

    # noinspection PyUnusedLocal
    @abstractmethod
    def __init__(
        self,
        item_types: Union[
            Iterable[Union["Property", type]], Types, None, "Property", type
        ] = None,
    ):
        pass

    @property  # type: ignore
    @abstractmethod
    def item_types(self) -> Optional[Types]:
        pass

    @item_types.setter  # type: ignore
    @abstractmethod
    def item_types(
        self,
        item_types: Union[
            Iterable[Union["Property", type]], Types, None, "Property", type
        ],
    ) -> None:
        pass


class Properties(Meta, metaclass=ABCMeta):

    # noinspection PyUnusedLocal
    @abstractmethod
    def __init__(
        self,
        items: Union[
            Mapping[str, "Property"],
            Iterable[Tuple[str, "Property"]],
            "Properties",
            None,
        ] = None,
    ) -> None:
        pass

    @abstractmethod
    def keys(self) -> KeysView[str]:
        pass

    @abstractmethod
    def values(self) -> ValuesView["Property"]:
        pass

    @abstractmethod
    def items(self) -> ItemsView[str, "Property"]:
        pass

    @abstractmethod
    def __setitem__(self, key: str, value: "Property") -> None:
        pass

    @abstractmethod
    def __copy__(self) -> "Properties":
        pass

    @abstractmethod
    def __deepcopy__(self, memo: dict = None) -> "Properties":
        pass

    @abstractmethod
    def __repr__(self) -> str:
        pass

    @abstractmethod
    def __getitem__(self, key: str) -> "Property":
        pass

    @abstractmethod
    def __delitem__(self, key: str) -> None:
        pass

    @abstractmethod
    def __iter__(self) -> Iterator[str]:
        pass

    @abstractmethod
    def __contains__(self, key: str) -> bool:
        pass

    @abstractmethod
    def pop(self, key: str, default: Undefined = UNDEFINED) -> "Property":
        pass

    @abstractmethod
    def popitem(self) -> Tuple[str, "Property"]:
        pass

    @abstractmethod
    def clear(self) -> None:
        pass

    @abstractmethod
    def update(
        self,
        *args: Union[
            Mapping[str, "Property"],
            Iterable[Tuple[str, "Property"]],
            "Properties",
        ],
        **kwargs: "Property",
    ) -> None:
        pass

    @abstractmethod
    def setdefault(
        self, key: str, default: Optional["Property"] = None
    ) -> Optional["Property"]:
        pass

    @abstractmethod
    def get(
        self, key: str, default: Optional["Property"] = None
    ) -> Optional["Property"]:
        pass

    @abstractmethod
    def __eq__(self, other: Any) -> bool:
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass


class Model(metaclass=ABCMeta):

    _format: Optional[str]
    _meta: Optional[Meta]
    _hooks: Optional[Hooks]
    _pointer: Optional[str]
    _url: Optional[str]

    @abstractmethod
    def _init_url(
        self,
        data: Union[
            Collection["MarshallableTypes"],
            Dict[str, "MarshallableTypes"],
            "Model",
            Readable,
            None,
        ],
    ) -> None:
        pass

    @abstractmethod
    def _init_format(
        self,
        data: Union[
            str,
            bytes,
            Readable,
            Mapping[str, "MarshallableTypes"],
            Iterable["MarshallableTypes"],
            "Model",
            None,
        ] = None,
    ) -> Union[
        Iterable["MarshallableTypes"],
        Mapping[str, "MarshallableTypes"],
        "Model",
        None,
    ]:
        pass

    @abstractmethod
    def _init_pointer(self) -> None:
        pass

    @abstractmethod
    def _marshal(self) -> "JSONTypes":
        pass

    @abstractmethod
    def _validate(self, raise_errors: bool = True) -> List[str]:
        pass


class Dictionary(Model, metaclass=ABCMeta):

    _meta: Optional[DictionaryMeta]

    @abstractmethod
    def __init__(
        self,
        items: Union[
            "Dictionary",
            Mapping[str, "MarshallableTypes"],
            Iterable[Tuple[str, "MarshallableTypes"]],
            Readable,
            str,
            bytes,
            None,
        ] = None,
        value_types: Union[
            Iterable[Union[type, "Property"]], type, "Property", Types, None
        ] = None,
    ) -> None:
        pass

    @abstractmethod
    def _marshal(self) -> OrderedDict[str, "JSONTypes"]:
        pass

    @abstractmethod
    def __setitem__(self, key: str, value: "MarshallableTypes") -> None:
        pass

    @abstractmethod
    def __delitem__(self, key: str) -> None:
        raise KeyError

    @abstractmethod
    def __getitem__(self, key: str) -> Any:
        pass

    @abstractmethod
    def __contains__(self, key: str) -> bool:
        pass

    @abstractmethod
    def __eq__(self, other: Any) -> bool:
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def __iter__(self) -> Iterator[str]:
        pass

    @abstractmethod
    def pop(self, key: str, default: Undefined = UNDEFINED) -> Any:
        pass

    @abstractmethod
    def popitem(self) -> Tuple[str, Any]:
        pass

    @abstractmethod
    def clear(self) -> None:
        pass

    @abstractmethod
    def update(
        self,
        *args: Union[
            Mapping[str, "MarshallableTypes"],
            Iterable[Tuple[str, "MarshallableTypes"]],
        ],
        **kwargs: "MarshallableTypes",
    ) -> None:
        pass

    @abstractmethod
    def setdefault(self, key: str, default: "MarshallableTypes" = None) -> Any:
        pass

    @abstractmethod
    def get(self, key: str, default: "MarshallableTypes" = None) -> Any:
        pass

    @abstractmethod
    def keys(self) -> KeysView[str]:
        pass

    @abstractmethod
    def items(self) -> ItemsView[str, Any]:
        pass

    @abstractmethod
    def values(self) -> ValuesView[Any]:
        pass


class Object(Model, metaclass=ABCMeta):
    """
    TODO
    """

    _meta: Optional[ObjectMeta]

    @abstractmethod
    def __init__(
        self,
        _data: Union[
            "Object",
            Dictionary,
            Mapping[str, "MarshallableTypes"],
            Iterable[Tuple[str, "MarshallableTypes"]],
            Readable,
            str,
            bytes,
            None,
        ] = None,
    ) -> None:
        pass

    @abstractmethod
    def _marshal(self) -> OrderedDict[str, "JSONTypes"]:
        pass

    @abstractmethod
    def __copy__(self) -> "Object":
        pass

    @abstractmethod
    def __deepcopy__(self, memo: Optional[dict]) -> "Object":
        pass

    @abstractmethod
    def __delattr__(self, key: str) -> None:
        pass

    @abstractmethod
    def __eq__(self, other: Any) -> bool:
        pass

    @abstractmethod
    def __getitem__(self, key: str) -> None:
        pass

    @abstractmethod
    def __hash__(self) -> int:
        pass

    @abstractmethod
    def __iter__(self) -> Iterable[str]:  # type: ignore
        pass

    @abstractmethod
    def __ne__(self, other: Any) -> bool:
        pass

    @abstractmethod
    def __repr__(self) -> str:
        pass

    @abstractmethod
    def __setattr__(
        self, property_name: str, value: "MarshallableTypes"
    ) -> None:
        pass

    @abstractmethod
    def __setitem__(self, key: str, value: "MarshallableTypes") -> None:
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass

    @abstractmethod
    def _get_property_validation_error_messages(
        self,
        property_name_: str,
        property_: "Property",
        value: "MarshallableTypes",
    ) -> Iterable[str]:
        pass

    def __reversed__(self) -> Iterator[str]:
        pass


class Array(Model, metaclass=ABCMeta):

    _meta: Optional[ArrayMeta]

    @abstractmethod
    def __init__(
        self,
        items: Union[
            "Array", Iterable["MarshallableTypes"], str, bytes, Readable
        ] = None,
        item_types: Union[
            Iterable[Union[type, "Property"]],
            type,
            "Property",
            Types,
            None,
        ] = None,
    ) -> None:
        pass

    @abstractmethod
    def _marshal(self) -> Sequence["JSONTypes"]:
        pass

    @abstractmethod
    def __getitem__(self, index: int) -> Any:
        pass

    @abstractmethod
    def __iter__(self) -> Iterator[Any]:
        pass

    @abstractmethod
    def __contains__(self, value: "MarshallableTypes") -> bool:
        pass

    @abstractmethod
    def __reversed__(self) -> "Array":
        pass

    @abstractmethod
    def index(
        self,
        value: "MarshallableTypes",
        start: int = 0,
        stop: Optional[int] = None,
    ) -> int:
        pass

    @abstractmethod
    def count(self, value: "MarshallableTypes") -> int:
        pass

    @abstractmethod
    def __setitem__(self, index: int, value: "MarshallableTypes") -> None:
        pass

    @abstractmethod
    def __delitem__(self, index: int) -> None:
        pass

    @abstractmethod
    def insert(self, index: int, value: "MarshallableTypes") -> None:
        pass

    @abstractmethod
    def append(self, value: "MarshallableTypes") -> None:
        pass

    @abstractmethod
    def clear(self) -> None:
        pass

    @abstractmethod
    def reverse(self) -> None:
        pass

    @abstractmethod
    def sort(
        self,
        key: Optional[Callable[[Any], Any]] = None,
        reverse: bool = False,
    ) -> None:
        pass

    @abstractmethod
    def extend(self, values: Iterable["MarshallableTypes"]) -> None:
        pass

    @abstractmethod
    def pop(self, index: int = -1) -> Any:
        pass

    @abstractmethod
    def remove(self, value: "MarshallableTypes") -> None:
        pass

    @abstractmethod
    def __iadd__(self, values: Iterable["MarshallableTypes"]) -> "Array":
        pass

    @abstractmethod
    def __add__(self, values: Iterable["MarshallableTypes"]) -> "Array":
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass


class Property(metaclass=ABCMeta):
    """
    TODO
    """

    name: Optional[str]
    required: bool
    _types: Optional[Types]

    # noinspection PyShadowingNames,PyUnusedLocal
    @abstractmethod
    def __init__(
        self,
        types: Iterable[Union[type, "Property"]] = None,
        name: Optional[str] = None,
        required: Union[bool, Callable] = False,
        versions: Optional[Iterable[Union[str, "Version"]]] = None,
    ) -> None:
        pass

    @property  # type: ignore
    @abstractmethod
    def types(self) -> Optional[Types]:
        pass

    @types.setter  # type: ignore
    @abstractmethod
    def types(
        self,
        types_or_properties: Optional[
            Union[Types, Sequence[Union[type, "Property"]], type, "Property"]
        ],
    ) -> None:
        pass

    @property  # type: ignore
    @abstractmethod
    def versions(self) -> Optional[Iterable["Version"]]:
        pass

    @versions.setter  # type: ignore
    @abstractmethod
    def versions(
        self,
        versions: Optional[
            Union[str, "Version", Iterable[Union[str, "Version"]]]
        ] = None,
    ) -> None:
        pass


class String(Property, metaclass=ABCMeta):
    """
    TODO
    """

    pass


class Date(Property, metaclass=ABCMeta):
    """
    TODO
    """

    @abstractmethod
    def date2str(self, value: date) -> str:
        pass

    @abstractmethod
    def str2date(self, value: str) -> date:
        pass


class DateTime(Property, metaclass=ABCMeta):
    @abstractmethod
    def datetime2str(self, value: datetime) -> str:
        pass

    @abstractmethod
    def str2datetime(self, value: str) -> datetime:
        pass


class Bytes(Property, metaclass=ABCMeta):
    """
    TODO
    """

    pass


class Enumerated(Property, metaclass=ABCMeta):
    """
    TODO
    """

    @property  # type: ignore
    def values(self) -> Optional[Set[Any]]:
        pass

    @values.setter
    def values(self, values: Optional[Iterable["MarshallableTypes"]]) -> None:
        pass


class Number(Property, metaclass=ABCMeta):

    pass


class Integer(Property, metaclass=ABCMeta):

    pass


class Boolean(Property, metaclass=ABCMeta):

    pass


class ArrayProperty(Property, metaclass=ABCMeta):
    """
    TODO
    """

    @property  # type: ignore
    @abstractmethod
    def item_types(self) -> Optional[Types]:
        pass

    @item_types.setter  # type: ignore
    @abstractmethod
    def item_types(
        self,
        item_types: Union[
            type, Property, Sequence[Union[type, Property]], Types, None
        ],
    ) -> None:
        pass


class DictionaryProperty(Property, metaclass=ABCMeta):
    """
    TODO
    """

    @property  # type: ignore
    @abstractmethod
    def value_types(self) -> Optional[Types]:
        pass

    @value_types.setter  # type: ignore
    @abstractmethod
    def value_types(
        self, value_types: Union[Sequence[Union[type, Property]], Types, None]
    ) -> None:
        pass


class Version(metaclass=ABCMeta):

    specification: Optional[str]
    equals: Optional[Sequence[Union[str, float, int, Decimal]]]
    not_equals: Optional[Sequence[Union[str, float, int, Decimal]]]
    less_than: Optional[Sequence[Union[str, float, int, Decimal]]]
    less_than_or_equal_to: Optional[Sequence[Union[str, float, int, Decimal]]]
    greater_than: Optional[Sequence[Union[str, float, int, Decimal]]]
    greater_than_or_equal_to: Optional[
        Sequence[Union[str, float, int, Decimal]]
    ]

    # noinspection PyUnusedLocal
    @abstractmethod
    def __init__(
        self,
        version_number: Optional[str] = None,
        specification: Optional[str] = None,
        equals: Optional[Sequence[Union[str, float, int, Decimal]]] = None,
        not_equals: Optional[Sequence[Union[str, float, int, Decimal]]] = None,
        less_than: Optional[Sequence[Union[str, float, int, Decimal]]] = None,
        less_than_or_equal_to: Optional[
            Sequence[Union[str, float, int, Decimal]]
        ] = None,
        greater_than: Optional[
            Sequence[Union[str, float, int, Decimal]]
        ] = None,
        greater_than_or_equal_to: Optional[
            Sequence[Union[str, float, int, Decimal]]
        ] = None,
    ) -> None:
        pass


MARSHALLABLE_TYPES: Tuple[type, ...] = (
    str,
    bytes,
    bytearray,
    bool,
    Mapping,
    Collection,
    Iterator,
    Model,
    int,
    float,
    decimal.Decimal,
    date,
    datetime,
    Null,
)
JSON_TYPES: Tuple[type, ...] = (
    str,
    int,
    float,
    bool,
    Mapping,
    Sequence,
    NoneType,
)
JSONTypes = Union[
    str,
    int,
    float,
    bool,
    Mapping[str, Any],
    Sequence,
    None,
]
MarshallableTypes = Union[
    bool,
    str,
    bytes,
    Model,
    Mapping[str, Any],
    Collection,
    Iterator,
    int,
    float,
    Decimal,
    date,
    datetime,
    Null,
    None,
]



import collections
from inspect import (
    FrameInfo,
    Parameter,
    getargvalues,
    getmodulename,
    getsource,
    signature,
    stack,
)
from types import ModuleType
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

import builtins
import sys

from .string import indent
from .types import UNDEFINED, Undefined


def properties_values(
    object_: object, include_private: bool = False
) -> Iterable[Tuple[str, Any]]:
    """
    This function iterates over an object's public (non-callable)
    properties, yielding a tuple comprised of each attribute/property name and
    value
    """
    for attribute in dir(object_):
        if include_private or attribute[0] != "_":
            value = getattr(object_, attribute, lambda: None)
            if not callable(value):
                yield attribute, value


QUALIFIED_NAME_ARGUMENT_TYPES: Tuple[Any, ...] = (
    type,
    collections.abc.Callable,
    ModuleType,
)


def qualified_name(type_or_module: Union[type, Callable, ModuleType]) -> str:
    """
    >>> print(qualified_name(qualified_name))
    sob.utilities.inspect.qualified_name

    >>> from sob import model
    >>> print(qualified_name(model.marshal))
    sob.model.marshal
    """
    assert isinstance(type_or_module, QUALIFIED_NAME_ARGUMENT_TYPES)
    type_name: str
    # noinspection SpellCheckingInspection
    if isinstance(type_or_module, ModuleType):
        type_name = type_or_module.__name__
    else:
        type_name = ".".join(
            name_part
            for name_part in getattr(
                type_or_module,
                "__qualname__",
                getattr(type_or_module, "__name__"),
            ).split(".")
            if name_part[0] != "<"
        )
        if type_or_module.__module__ not in (
            "builtins",
            "__builtin__",
            "__main__",
            "__init__",
        ):
            type_name = type_or_module.__module__ + "." + type_name
    return type_name


def calling_functions_qualified_names(depth: int = 1) -> List[str]:
    """
    This function returns the qualified names of all calling functions in the
    stack, starting with the function at the indicated `depth` (defaults to 1).

    >>> def my_function_a():
    ...     return calling_functions_qualified_names()
    >>> def my_function_b():
    ...     return my_function_a()
    >>> print('\\n'.join(my_function_b()[-2:]))
    sob.utilities.inspect.calling_functions_qualified_names.my_function_b
    sob.utilities.inspect.calling_functions_qualified_names.my_function_a
    """
    depth += 1
    name = calling_function_qualified_name(depth=depth)
    names: List[str] = []
    while name:
        if name and not (names and names[0] == name):
            names.insert(0, name)
        depth += 1
        name = calling_function_qualified_name(depth=depth)
    return names


def _get_module_name(file_name: str) -> str:
    """
    Given a frame info's file name, find the module name
    """
    module_name = getmodulename(file_name)
    if module_name is None:
        # Check to see if this is a doctest
        doc_test_prefix: str = "<doctest "
        if file_name.startswith(doc_test_prefix):
            doc_test_prefix_length: int = len(doc_test_prefix)
            module_name = file_name[doc_test_prefix_length:]
            module_name = module_name.rstrip(">")
            if "[" in module_name:
                module_name = "[".join(module_name.split("[")[:-1])
        else:
            raise ValueError(
                'The path "%s" is not a python module' % file_name
            )
    return module_name


def _get_frame_info_names(frame_info: FrameInfo) -> List[str]:
    names: List[str] = []
    if frame_info.function != "<module>":
        names.append(frame_info.function)
        arguments, _, _, frame_locals = getargvalues(frame_info.frame)
        if arguments:
            argument = arguments[0]
            argument_value = frame_locals[argument]
            argument_value_type = type(argument_value)
            if (
                hasattr(argument_value_type, "__name__")
                and hasattr(argument_value_type, "__module__")
                and (
                    (argument_value_type.__name__ not in builtins.__dict__)
                    or (
                        builtins.__dict__[argument_value_type.__name__]
                        is not argument_value_type
                    )
                )
            ):
                names.append(qualified_name(argument_value_type))
    if len(names) < 2:
        module_name = _get_module_name(frame_info.filename)
        if module_name in sys.modules:
            qualified_module_name = qualified_name(sys.modules[module_name])
            names.append(qualified_module_name)
        elif module_name:
            names.append(module_name)
    return names


def calling_module_name(depth: int = 1) -> str:
    """
    This function returns the name of the module from which the function
    which invokes this function was called.

    Parameters:

    - depth (int): This defaults to `1`, indicating we want to return the name
      of the module wherein `calling_module_name` is being called. If set to
      `2`, it would instead indicate the module

    >>> print(calling_module_name())
    sob.utilities.inspect

    >>> print(calling_module_name(2))
    doctest
    """
    name: str
    try:
        name = getattr(sys, "_getframe")(depth).f_globals.get(
            "__name__", "__main__"
        )
    except (AttributeError, ValueError):
        name = "__main__"
    return name


# noinspection PyUnresolvedReferences
def calling_function_qualified_name(depth: int = 1) -> Optional[str]:
    """
    Return the fully qualified name of the function from within which this is
    being called

    >>> def my_function():
    ...     return calling_function_qualified_name()
    >>> print(my_function())
    sob.utilities.inspect.calling_function_qualified_name.my_function

    >>> class MyClass:
    ...
    ...     def __call__(self) -> None:
    ...          return self.my_method()
    ...
    ...     # noinspection PyMethodMayBeStatic
    ...     def my_method(self) -> str:
    ...          return calling_function_qualified_name()
    >>> print(MyClass()())
    sob.utilities.inspect.MyClass.my_method
    """
    assert isinstance(depth, int)
    try:
        stack_ = stack()
    except IndexError:
        return None
    if len(stack_) < (depth + 1):
        return None
    names: List[str] = _get_frame_info_names(stack_[depth])
    return ".".join(reversed(names))


def get_source(object_: Union[type, Callable, ModuleType]) -> str:
    """
    Get the source code which defined an object.
    """
    object_source: str = getattr(object_, "_source", "")
    if not object_source:
        object_source = getsource(object_)
    return object_source


def parameters_defaults(function: Callable[..., Any]) -> Dict[str, Any]:
    """
    Returns an ordered dictionary mapping a function's argument names to
    default values, or `UNDEFINED` in the case of
    positional arguments.

    >>> class X:
    ...
    ...    def __init__(self, a, b, c, d=1, e=2, f=3):
    ...        pass
    ...
    >>> for parameter_name, default in parameters_defaults(X.__init__).items():
    ...     print((parameter_name, default))
    ('self', UNDEFINED)
    ('a', UNDEFINED)
    ('b', UNDEFINED)
    ('c', UNDEFINED)
    ('d', 1)
    ('e', 2)
    ('f', 3)
    """
    defaults: Dict[str, Any] = collections.OrderedDict()
    parameter_name: str
    parameter: Parameter
    for parameter_name, parameter in signature(function).parameters.items():
        if parameter.default is Parameter.empty:
            defaults[parameter_name] = UNDEFINED
        else:
            defaults[parameter_name] = parameter.default
    return defaults


def _repr_items(items: Union[Sequence, Set]) -> str:
    """
    Returns a string representation of the items in a `list`, `tuple`, or
    `set`.
    """
    lines: List[str] = []
    for item in items:
        lines.append(indent(represent(item), start=0))
    return ",\n".join(lines)


def _repr_list(list_instance: list) -> str:
    """
    Returns a string representation of `list` argument values
    """
    return "[\n%s\n]" % _repr_items(list_instance)


def _repr_tuple(tuple_instance: tuple) -> str:
    """
    Returns a string representation of `tuple` argument values
    """
    return "(\n{}\n)".format(
        _repr_items(tuple_instance) + ("," if len(tuple_instance) == 1 else "")
    )


def _repr_set(set_instance: set) -> str:
    """
    Returns a string representation of `set` argument values
    """
    return "{\n%s\n}" % _repr_items(
        sorted(set_instance, key=lambda item: repr(item))
    )


def represent(value: Any) -> str:
    """
    Returns a string representation of a value.
    """
    value_representation: str
    if isinstance(value, type):
        value_representation = qualified_name(value)
    else:
        value_type: type = type(value)
        if value_type is list:
            assert isinstance(value, list)
            value_representation = _repr_list(value)
        elif value_type is tuple:
            assert isinstance(value, tuple)
            value_representation = _repr_tuple(value)
        elif value_type is set:
            assert isinstance(value, set)
            value_representation = _repr_set(value)
        else:
            value_representation = repr(value)
    return value_representation


def get_method(
    object_instance: object,
    method_name: str,
    default: Union[Callable, Undefined, None] = UNDEFINED,
) -> Optional[Callable[..., Any]]:
    """
    This function attempts to retrieve a method, by name.

    Parameters:

    - object_instance (object)
    - method_name (str)
    - default (collections.Callable|None) = None

    This function returns an object's method, if the method exists.
    If the object does not have a method with the given name, this
    function returns `None`.
    """
    method: Callable
    try:
        method = getattr(object_instance, method_name)
    except AttributeError:
        if isinstance(default, Undefined):
            raise
        else:
            return default
    if callable(method):
        return method
    else:
        if isinstance(default, Undefined):
            raise AttributeError(
                f"{qualified_name(type(object_instance))}.{method_name} "
                "is not callable."
            )
        else:
            return method
